# Web browsers send an Accept request header for initial (e.g. non-AJAX) page requests
# which should contain 'text/html'
MIME_TYPE_HTML = 'text/html'
MIME_TYPE_JSON = 'application/json'
MIME_TYPE_LD_JSON = 'application/ld+json'
